@@ -39,13 +39,9 @@
 #include <libgnomecups/gnome-cups-ui-connection.h>
 #include <libgnomecups/gnome-cups-permission.h>
 
+#include <gtk/gtkaccessible.h>
 #include <cups/cups.h>
 
-typedef struct {
-	GnomeCupsPrinter *printer;
-	char *value;
-} PrinterInfoData;
-
 static void
 unref_group (gpointer user_data, GObject *obj)
 {
@@ -103,7 +99,8 @@ update_general_page (GladeXML *xml)
 	w = glade_xml_get_widget (xml, "printer_image");
 	gtk_image_set_from_pixbuf (GTK_IMAGE (w), pixbuf);
 
-	label_vs_entry (xml, "name_entry", "name_label",
+	w = glade_xml_get_widget (xml, "name_label");
+	gtk_label_set_text (GTK_LABEL (w),
 		gnome_cups_printer_get_name (printer));
 	label_vs_entry (xml, "description_entry", "description_label",
 		gnome_cups_printer_get_description (printer));
@@ -115,14 +112,24 @@ update_general_page (GladeXML *xml)
 		gnome_cups_printer_get_full_state (printer));
 }
 
+typedef struct {
+	GnomeCupsPrinter *printer;
+	char *value;
+	GtkWidget *entry;
+} EntryHandlerInfo;
+
 static gboolean
-apply_description_handler (gpointer user_data)
+cb_entry_idle_handler (EntryHandlerInfo *data)
 {
-	PrinterInfoData *data = user_data;
+	gchar const *name = gtk_widget_get_name (data->entry);
+
+	g_return_val_if_fail (name != NULL, FALSE);
+
+	if (0 == strcmp (name, "description_entry"))
+		gnome_cups_printer_set_description (data->printer, data->value, NULL);
+	else if (0 == strcmp (name, "location_entry"))
+		 gnome_cups_printer_set_location (data->printer, data->value, NULL);
 
-	gnome_cups_printer_set_description (data->printer,
-					    data->value,
-					    NULL);
 	g_object_unref (data->printer);
 	g_free (data->value);
 	g_free (data);
@@ -130,79 +137,35 @@ apply_description_handler (gpointer user_data)
 	return FALSE;
 }
 
-static void
-apply_description (GladeXML *xml, GtkWidget *entry)
+/* You can't hit the main loop in a GtkEntry's focus_out_event, so queue this
+ * in an idle handler */
+static gboolean
+cb_entry_activated (GtkEntry *entry, GladeXML *xml)
 {
-	GnomeCupsPrinter *printer;
-	PrinterInfoData *data;
-
-	/* You can't hit the main loop in a GtkEntry's focus_out_event,
-	 * so queue this in an idle handler */
-
-	data = g_new0 (PrinterInfoData, 1);
-	printer = g_object_get_data (G_OBJECT (xml), "printer");
+	EntryHandlerInfo *data = g_new0 (EntryHandlerInfo, 1);
+	GnomeCupsPrinter *printer = g_object_get_data (G_OBJECT (xml), "printer");
 	g_object_ref (printer);
 	data->printer = printer;
-	data->value = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry)));
-
-	g_idle_add (apply_description_handler, data);
-}
-
-static gboolean
-apply_location_handler (gpointer user_data)
-{
-	PrinterInfoData *data = user_data;
-
-	gnome_cups_printer_set_location (data->printer,
-					 data->value,
-					 NULL);
-	g_object_unref (data->printer);
-	g_free (data->value);
-	g_free (data);
-
+	data->value = g_strdup (gtk_entry_get_text (entry));
+	data->entry = GTK_WIDGET (entry);
+	g_idle_add ((GSourceFunc)cb_entry_idle_handler, data);
 	return FALSE;
 }
+static gboolean
+cb_entry_lost_focus (GtkEntry *entry, G_GNUC_UNUSED GdkEventFocus *event,
+		     GladeXML *xml)
+{
+	return cb_entry_activated (entry, xml);
+}
 
 static void
-apply_location (GladeXML *xml, GtkWidget *entry)
+setup_entry (GladeXML *xml, char const *name)
 {
-	GnomeCupsPrinter *printer;
-	PrinterInfoData *data;
-
-	/* You can't hit the main loop in a GtkEntry's focus_out_event,
-	 * so queue this in an idle handler */
-
-	data = g_new0 (PrinterInfoData, 1);
-	printer = g_object_get_data (G_OBJECT (xml), "printer");
-	g_object_ref (printer);
-	data->printer = printer;
-	data->value = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry)));
-
-	g_idle_add (apply_location_handler, data);
-}
-
-static gboolean
-description_focus_out_cb (GtkWidget *entry,
-			  GdkEventFocus *event,
-			  gpointer user_data)
-{
-	GladeXML *xml = GLADE_XML (user_data);
-
-	apply_description (xml, entry);
-
-	return FALSE;
-}
-
-static gboolean
-location_focus_out_cb (GtkWidget *entry,
-		       GdkEventFocus *event,
-		       gpointer user_data)
-{
-	GladeXML *xml = GLADE_XML (user_data);
-
-	apply_location (xml, entry);
-
-	return FALSE;
+	GtkWidget *w = glade_xml_get_widget (xml, name);
+	g_signal_connect (w, "focus_out_event",
+		G_CALLBACK (cb_entry_lost_focus), xml);
+	g_signal_connect (w, "activate",
+		G_CALLBACK (cb_entry_activated), xml);
 }
 
 static void
@@ -386,20 +349,11 @@ test_page_clicked_cb (GtkWidget *button, gpointer user_data)
 static void
 setup_general_page (GladeXML *xml)
 {
-	GtkWidget *widget;
-
 	update_general_page (xml);
 
 	if (gnome_cups_can_admin ()) {
-		widget = glade_xml_get_widget (xml, "description_entry");
-		g_signal_connect (widget,
-			"focus_out_event",
-			G_CALLBACK (description_focus_out_cb), xml);
-
-		widget = glade_xml_get_widget (xml, "location_entry");
-		g_signal_connect (widget,
-			"focus_out_event",
-			G_CALLBACK (location_focus_out_cb), xml);
+		setup_entry (xml, "description_entry");
+		setup_entry (xml, "location_entry");
 	}
 
 	hookup_option_menu_xml (xml, "resolution_option_menu", "Resolution");
@@ -483,6 +437,43 @@ printer_gone_cb (GnomeCupsPrinter *printer, gpointer user_data)
 	close_window (xml);
 }
 
+static void
+set_atk_relation (GtkWidget *label, GtkWidget *widget)
+{
+	AtkObject *atk_widget;
+	AtkObject *atk_label;
+	AtkRelationSet *relation_set;
+	AtkRelation *relation;
+	AtkObject *targets[1];
+
+	atk_widget = gtk_widget_get_accessible (widget);
+	atk_label = gtk_widget_get_accessible (label);
+
+	/* Create the label -> widget relation */
+	targets[0] = atk_widget;
+	relation_set = atk_object_ref_relation_set (atk_label);
+	relation = atk_relation_new (targets, 1, ATK_RELATION_LABEL_FOR);
+	atk_relation_set_add (relation_set, relation);
+	g_object_unref (relation);
+	g_object_unref (relation_set);
+
+	/* Create the widget -> label relation */
+	targets[0] = atk_label;
+	relation_set = atk_object_ref_relation_set (atk_widget);
+	relation = atk_relation_new (targets, 1, ATK_RELATION_LABELLED_BY);
+	atk_relation_set_add (relation_set, relation);
+	g_object_unref (relation);
+	g_object_unref (relation_set);
+}
+
+static void
+printer_attributes_changed_cb (GnomeCupsPrinter *printer, gpointer user_data)
+{
+	GladeXML *xml = GLADE_XML (user_data);
+
+	update_general_page (xml);
+}
+
 static GtkWidget *
 create_option_editor (GladeXML *xml,
 		      GnomeCupsPrinter *printer,
@@ -510,6 +501,7 @@ create_option_editor (GladeXML *xml,
 		gtk_widget_show (check);
 		gtk_box_pack_start (GTK_BOX (hbox), check,
 				    FALSE, FALSE, 0);
+		set_atk_relation (label, check);
 		hookup_check_button (xml, printer, check, option->id);
 	} else if (option->type == GNOME_CUPS_PRINTER_OPTION_PICK_ONE) {
 		GtkWidget *option_menu;
@@ -519,6 +511,7 @@ create_option_editor (GladeXML *xml,
 		gtk_size_group_add_widget (editor_group, option_menu);
 		gtk_box_pack_start (GTK_BOX (hbox), option_menu,
 				    FALSE, FALSE, 0);
+		set_atk_relation (label, option_menu);
 		hookup_option_menu (xml, printer, option_menu, option->id);
 	}
 
@@ -630,7 +623,7 @@ cb_connection_changed (GCupsConnectionSelector *cs, GnomeCupsPrinter *printer)
 	if (uri != NULL) {
 		ipp_t *request = gnome_cups_request_new_for_printer (
 			CUPS_ADD_PRINTER, printer);
-		ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_NAME,
+		ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_URI,
 			"device-uri", NULL, gnome_cups_strdup (uri));
 		gnome_cups_request_execute_async (request, NULL, "/admin/", NULL, NULL, NULL);
 		g_free (uri);
@@ -671,7 +664,7 @@ setup_connection_page (GladeXML *xml, GnomeCupsPrinter *printer)
 	gtk_widget_show (cs);
 	g_signal_connect_object (cs,
 		"changed",
-		G_CALLBACK (cb_connection_changed), xml, 0);
+		G_CALLBACK (cb_connection_changed), printer, 0);
 }
 
 static void
@@ -711,6 +704,8 @@ gnome_cups_manager_printer_properties (GnomeCupsPrinter *printer)
 				printer, g_object_unref);
 	g_signal_connect_object (printer, "gone",
 				 G_CALLBACK (printer_gone_cb), xml, 0);
+	g_signal_connect_object (printer, "attributes_changed",
+				 G_CALLBACK (printer_attributes_changed_cb), xml, 0);
 
 	g_object_weak_ref (G_OBJECT (xml), disconnect_view, printer);
 
